function defineMiddleware(middleware) {
  return middleware;
}

class FatcherError extends Error {
  constructor(request, response) {
    super(`[fatcher] Request fail with code ${response.status}`);
    this.snapshot = {
      request,
      response
    };
  }
  name = "fatcherError";
  snapshot;
}

const exception = (options = {}) => {
  return defineMiddleware(async (context, next) => {
    const { validateCode } = options;
    const response = await next();
    if (validateCode?.(response.status)) {
      return response;
    }
    if (response.ok) {
      return response;
    }
    return Promise.reject(new FatcherError(context, response));
  });
};

function merge(target, source, source1) {
  if (!source) {
    return target;
  }
  for (const key of Object.keys(source)) {
    if (!Object.hasOwn(Object.getPrototypeOf(target), key)) {
      target[key] = source[key];
    }
  }
  if (source1) {
    return merge(target, source1);
  }
  return target;
}

function composeMiddlewares(middlewares) {
  return function use(init) {
    let currentIndex = -1;
    let response;
    let request = init;
    async function dispatch(index, patch) {
      if (index <= currentIndex) {
        return response;
      }
      currentIndex = index;
      const middleware = middlewares[index];
      if (!middleware) {
        return response;
      }
      if (patch) {
        const base = patch.url ? new Request(patch.url, request) : request;
        request = merge(new Request(base, patch), { ...request }, patch);
      }
      const newResponse = await middleware(request, async (_) => dispatch(index + 1, _));
      if (response !== newResponse) {
        response = merge(newResponse, response);
      } else {
        response = newResponse;
      }
      return response;
    }
    return dispatch(0);
  };
}

async function fatcher(input, options = {}) {
  const { middlewares = [], ...rest } = options;
  const context = new Request(input, rest);
  context.options = rest;
  let middlewareList = [];
  for (const middleware of middlewares) {
    if (Array.isArray(middleware)) {
      middlewareList = middlewareList.concat(middleware);
    } else {
      middlewareList.push(middleware);
    }
  }
  return composeMiddlewares([...middlewareList, (req) => fetch(req)])(context);
}

function isFatcherError(error) {
  return error instanceof FatcherError && error.name === "fatcherError";
}

export { FatcherError, defineMiddleware, exception, fatcher, isFatcherError };
